// src/entry.rs

static READY: spin::Once<()> = spin::Once::new();

/// After some initialization in asm/entry.S, the kernel will jump here and
/// each hart will have its own setup sequence.
#[no_mangle]
unsafe extern "C" fn _start() {
    use core::arch::asm;
    use riscv::register::{mepc, mstatus, satp, sie, sstatus};

    // Set previous privilege for all harts to M-mode, set previous interrupt
    // enable, and set floating-point unit to initial state
    mstatus::set_mpp(mstatus::MPP::Supervisor);
    mstatus::set_mpie();
    mstatus::set_fs(sstatus::FS::Initial);

    // Store ref to main in M-mode exception program counter
    mepc::write(main as usize);

    // Disable paging
    satp::write(0);

    // Delegate all traps to S-mode, using inline assembly since we have not
    // provided a wrapper for it yet
    asm!("li t0, 0xffff", "csrw medeleg, t0", "csrw mideleg, t0",);

    // Enable S-mode external, timer, and software interrupts
    sie::set_sext();
    sie::set_stimer();
    sie::set_ssoft();

    // TODO configure PMP
    // TODO timer init

    // Use mret to jump to main
    asm!("mret");
}

extern "C" fn main() {
    use riscv::interrupt;
    use riscv::register::{mstatus, sstatus};

    if hartid() == 0 {
        READY.call_once(|| {
            // Disable machine interrupts while initializing
            interrupt::machine::disable();
            // TODO Initialize console
            // TODO Write boot message

            // TODO Set up paging
            // TODO Set up processes

            // TODO Set up trap vectors
            // TODO Set up PLIC

            kinit();
            unsafe {
                mstatus::set_mpp(mstatus::MPP::User);
                mstatus::set_mpie();
                mstatus::set_spie();
                mstatus::set_fs(sstatus::FS::Initial);
            }

            riscv::asm::fence();
        });
    } else {
        READY.wait();
        kinit_hart();
    }
}

#[no_mangle]
extern "C" fn kinit() {
    use crate::uart;
    uart::Device::new(0x1000_0000);
}

#[no_mangle]
extern "C" fn kinit_hart() {}

#[inline]
fn hartid() -> usize {
    use core::arch::asm;
    let id: usize;
    unsafe {
        asm!(
            "mv {id}, tp",
            id = out(reg) id,
        );
    }
    id
}
